'''
@author: Christian Dewey
@date: Jul 26, 2024
'''
from pandas import DataFrame
from lcicpms.in_out import get_data_from_csv

class RawICPMSData:
    """
    A class to represent and process raw ICP-MS data from a file.

    Attributes:
        raw_data_file (str): The file path of the raw ICP-MS data.
        data_type (str): The type of the raw data file (e.g., 'csv').
        min_time (float): The minimum time value in the separation.
        max_time (float): The maximum time value in the separation.
        intensities (dict): A dictionary containing intensities for each element, with elements as keys.
        times (dict): A dictionary containing time points for each element, with elements as keys.
        elements (list): A list of elements in the raw file.
        time_labels (list): A list of time labels extracted from the raw file.
        raw_data_df (DataFrame): A pandas DataFrame containing the raw data from the file.
    """
    
    def __init__(self, raw_data_file: str = None):
        """
        Initializes the RawICPMSData object by loading the data from the provided file.

        Args:
            raw_data_file (str, optional): The file path to the raw ICP-MS data file. Defaults to None.
        """
        # name of raw ICPMS file 
        self.raw_data_file = raw_data_file
        # str describing raw file data type; defined when data is loaded
        self.data_type: str = None
        # min time in the separation
        self.min_time: float = -1.0
        # max time in the separation 
        self.max_time: float = -1.0
        # dictionary of arrays containing intensities collected through time, elements are keys 
        self.intensities: dict = {}
        # dictionary of arrays containing time point of intensity measurement, elements are keys
        self.times: dict = {}
        # list of elements in raw file
        self.elements: list = []
        # list of time labels 
        self.time_labels: list = []
        # dataframe containing raw data 
        self.raw_data_df: DataFrame = None

        self.load_data()
        self.get_elements_and_time_labels()
        self.get_intensities()
        self.get_times()

    def load_data(self):
        """
        Determines the data type of the raw data file and loads the data.
        Currently, only .csv files are supported.
        """
        if self.raw_data_file.split('.')[-1] == 'csv':
            self.data_type = 'csv'
            self.raw_data_df = get_data_from_csv(self.raw_data_file)

    def get_elements_and_time_labels(self):
        """
        Extracts the elements and time labels from the raw data file, depending on the data type.
        The elements and time labels are identified by examining column names.
        """
        if self.data_type == 'csv':
            for col in self.raw_data_df.columns:
                if ('Time' not in col) and ('time' not in col) and ('Number' not in col):
                    self.elements.append(col)
                elif ('Time' in col) or ('time' in col):
                    self.time_labels.append(col)

    def get_intensities(self):
        """
        Extracts the intensities of each element from the raw data and stores them in a dictionary.
        The intensities are converted to NumPy arrays for easier processing.
        """
        self.intensities = {e: self.raw_data_df[e].to_numpy() for e in self.elements}

    def get_times(self):
        """
        Extracts the time points corresponding to each element and stores them in a dictionary.
        The time labels are used to map the correct time values to each element.
        """
        time_labels_dict = {e: [l for l in self.time_labels if e in l][0] for e in self.elements}
        self.times = {e: self.raw_data_df[time_labels_dict[e]].to_numpy() for e in self.elements}

    def plot_raw_data(self, elements: list = None):
        """
        Plots the raw ICP-MS data for the specified elements using matplotlib.

        Args:
            elements (list, optional): A list of elements to plot. If None, all elements are plotted.
        """
        import matplotlib.pyplot as plt

        if elements is None:
            elements = self.elements

        fig, ax = plt.subplots()
        maxy = 0
        for e in elements:
            ax.plot(self.times[e], self.intensities[e], label=e)
            if max(self.intensities[e]) > maxy:
                maxy = max(self.intensities[e])
                ax.set_ylim(0, maxy * 1.01)
        ax.legend(frameon=False)
        title = self.raw_data_file.split('/')[-1]
        fig.suptitle(title)
        plt.show()


'''
@author: Christian Dewey
@date: September 4, 2024
'''

import os 
from pandas import concat

from lcicpms.raw_icpms_data import RawICPMSData
from lcicpms.integrate import Integrate
from lcicpms.quantitate import Quantitate
from lcicpms.calibration import Calibration


class Dataset:
    """
    A class to handle raw LC-ICP-MS data, perform calibration, 
    internal standard corrections, and quantify element concentrations.
    
    Attributes:
        raw_data_dir (str): Directory containing raw ICP-MS data files.
        cal_data_dir (str): Directory containing calibration files.
        skip_keywords (list): List of keywords to exclude certain files.
        raw_data_dict (dict): Dictionary of loaded raw data files.
        cal_icpms_obj_dict (dict): Dictionary of calibration files.
        _cal_has_run (bool): Flag indicating whether calibration has been performed.
        cal (Calibration): Calibration object containing standard concentrations and elements.
        concentrations_df (DataFrame): Final concentrations of elements after quantification.
    """
    
    def __init__(self, raw_data_dir: str = None, 
                 cal_data_dir: str = None,
                 skip_keywords: list = ['results']):
        """
        Initialize the Dataset object with directories for raw and calibration data.
        
        Args:
            raw_data_dir (str): Directory containing raw data files.
            cal_data_dir (str): Directory containing calibration data files.
            skip_keywords (list): List of keywords to filter out unwanted files (default: ['results']).
        """
        self.raw_data_dir = raw_data_dir
        self.skip_keywords = skip_keywords
        self._cal_has_run = False
        self.raw_data_dict = self.load_raw_data(self.raw_data_dir)
        
        if cal_data_dir is None:
            self.cal_data_dir = raw_data_dir
        else:
            self.cal_data_dir = cal_data_dir


    def load_raw_data(self, dir):
        """
        Load raw ICP-MS data files from the specified directory.
        
        Args:
            dir (str): Directory to load raw data from.
        
        Returns:
            dict: Dictionary where keys are file paths and values are RawICPMSData objects.
        """
        print(f'loading all files in {dir}')
        flist = [os.path.join(dir, f)
                 for f in os.listdir(dir) 
                 if f.endswith('.csv') and not any(s in f for s in self.skip_keywords)]
        
        icpms_obj_dict = {}
        for f in flist:
            icpms_obj = RawICPMSData(f)
            icpms_obj_dict[f] = icpms_obj
        
        self.elements = icpms_obj.elements
        return icpms_obj_dict
    
    
    def run_calibration(self,
                        cal_std_concs: list = [0, 10, 25, 50, 100, 200], 
                        cal_keywords_by_conc: list = ['std_0', 'std_1', 'std_2', 'std_3', 'std_4', 'std_5']):
        """
        Perform calibration using known standard concentrations and associated files.
        
        Args:
            cal_std_concs (list): List of standard concentrations (default: [0, 10, 25, 50, 100, 200]).
            cal_keywords_by_conc (list): Keywords to identify files for each concentration (default: ['std_0', 'std_1', 'std_2', 'std_3', 'std_4', 'std_5']).
        """
        print(f'loading calibration files in {self.cal_data_dir}')
        flist = [os.path.join(self.cal_data_dir, f)
                 for f in os.listdir(self.cal_data_dir) 
                 if f.endswith('.csv') and not any(s in f for s in self.skip_keywords)]
        
        cal_files = [f for f in flist if any(s in f for s in cal_keywords_by_conc)]
        self.cal_icpms_obj_dict = {k: None for k in cal_keywords_by_conc}

        for k in cal_keywords_by_conc:
            for f in cal_files:
                if k in f:
                    cal_f = f
            self.cal_icpms_obj_dict[k] = cal_f
        
        ordered_cal_files = [os.path.join(self.cal_data_dir, self.cal_icpms_obj_dict[c]) for c in cal_keywords_by_conc]
        self._cal_has_run = True
        
        # Create Calibration object
        self.cal = Calibration(concentrations=cal_std_concs, elements=self.elements, rawfiles=ordered_cal_files)


    def internal_std_correction(self):
        """
        Perform internal standard correction using the element "115In".
        
        Returns:
            float: Baseline value for the internal standard "115In".
        
        Raises:
            Exception: If calibration hasn't been run or if "115In" is not found in the data.
        """
        if not self._cal_has_run:
            raise Exception("Run calibration data before internal standard correction!")
        
        print('running 115In internal standard correction')
        stds_list = list(self.cal_icpms_obj_dict.keys())
        icpms_obj = RawICPMSData(os.path.join(self.cal_data_dir, self.cal_icpms_obj_dict[stds_list[0]]))
        
        if '115In' not in icpms_obj.intensities:
            raise Exception("115In was not found in the raw data file. Cannot perform internal standard correction.")
        
        istd_base = Integrate.integrate(icpms_obj.intensities['115In'], icpms_obj.times['115In'])
        return istd_base
    

    def quantitate(self,
                   time_range: tuple = (-1, -1),
                   cal_std_concs: list = [0, 10, 25, 50, 100, 200], 
                   cal_keywords_by_conc: list = ['std_0', 'std_1', 'std_2', 'std_3', 'std_4', 'std_5']):
        """
        Quantify concentrations of all elements in raw file using calibration data and internal standard correction.
        
        Args:
            time_range (tuple): Tuple indicating time range for concentration integration (default: (-1, -1)).
            cal_std_concs (list): List of standard concentrations for calibration (default: [0, 10, 25, 50, 100, 200]).
            cal_keywords_by_conc (list): Keywords to identify calibration files (default: ['std_0', 'std_1', 'std_2', 'std_3', 'std_4', 'std_5']).
        
        Returns:
            DataFrame: DataFrame containing quantified concentrations for each element in each file.
        """
        if not self._cal_has_run:
            cal = self.run_calibration(cal_std_concs=cal_std_concs, cal_keywords_by_conc=cal_keywords_by_conc)
        else:
            cal = self.cal

        istd_base = self.internal_std_correction()

        conc_list = []
        for file, data in self.raw_data_dict.items():
            
            tmin = 'min time' if time_range[0] == -1 else time_range[0]
            tmax = 'max time' if time_range[1] == -1 else time_range[1]
                
            print(f'integrating from {tmin} to {tmax} for {file}')
            
            t_start, t_stop = time_range

            conc = Quantitate.run(data,
                                  cal,
                                  data.elements,
                                  time_range=(t_start, t_stop),
                                  istd_baseline=istd_base,
                                  istd_time_range=(60, 120))
            
            conc_df = conc['concentrations']
            conc_df['file'] = file.split('/')[-1]
            conc_df['int_std_correction'] = conc['internal_std_correction']
            conc_list.append(conc_df)

        self.concentrations_df = concat(conc_list)



'''
@author: Christian Dewey
@date: Jul 27, 2024
'''
from lcicpms.integrate import Integrate
from lcicpms.raw_icpms_data import RawICPMSData

from sklearn import linear_model
from sklearn.metrics import mean_squared_error, r2_score
from numpy import array

class Curve:
    """
    A class to represent a calibration curve for a specific element.

    Attributes:
        element (str): The element for which the curve is created.
        concentrations (list): A list to store concentrations of the element.
        peak_areas (list): A list to store the integrated peak areas for the element.
        lm (LinearRegression): Linear regression model for the calibration curve.
        r2 (float): The R-squared value of the regression.
        mse (float): The mean squared error of the regression.
        m (float): The slope of the calibration curve.
        b (float): The intercept of the calibration curve.
        prediction (ndarray): The predicted concentrations based on the regression.
    """
    def __init__(self, element):
        """
        Initializes the Curve instance with the specified element and sets up empty 
        lists for concentrations and peak areas.
        
        Args:
            element (str): The name of the element for which the curve is being created.
        """
        self.element = element
        self.concentrations = []
        self.peak_areas = []
        self.lm = None
        self.r2 = None
        self.mse = None
        self.m = None
        self.b = None
        self.prediction = None


class Calibration:
    """
    A class to perform calibration of elements using LC-ICP-MS data. It generates 
    calibration curves and fits a linear regression model for each element.

    Attributes:
        elements (list): A list of elements to calibrate.
        rawfiles (list): A list of raw ICP-MS data files.
        curves (dict): A dictionary of Curve objects for each element.
    """
    def __init__(self, concentrations: list | dict, elements: list, rawfiles: list):
        """
        Initializes the Calibration instance and builds calibration curves for the elements.

        Args:
            concentrations (list | dict): A list or dictionary of known concentrations for calibration.
            elements (list): A list of element names to calibrate.
            rawfiles (list): A list of raw ICP-MS data files corresponding to the concentrations.
        """
        self.elements = elements
        self.rawfiles = rawfiles
        self.curves = {e: Curve(e) for e in elements}

        self.build_curve(concentrations)
        self.run_regression()

    def build_curve(self, concentrations: list | dict):
        """
        Builds the calibration curve for each element by integrating the peak areas from the raw data.

        Args:
            concentrations (list | dict): A list or dictionary of concentrations mapped to raw data files.
        """
        rawfiles = self.rawfiles

        if type(concentrations) == list:
            elements = self.elements
            concentrations = {e: {c: file for c, file in zip(concentrations, rawfiles)} for e in elements}

        elif type(concentrations) == dict:
            elements = list(concentrations.keys())
            concentrations = {e: {c: file for c, file in zip(concentrations, rawfiles)} for e in elements}

        for e in elements:
            for c, f in concentrations[e].items():
                raw = RawICPMSData(f)
                peak_area = Integrate.integrate(raw.intensities[e], raw.times[e])
                self.curves[e].concentrations.append(c)
                self.curves[e].peak_areas.append(peak_area)

    def plot_curves(self, ax=None):
        """
        Plots the calibration curves for each element using peak areas and concentrations.

        Args:
            ax (matplotlib.axes.Axes, optional): A matplotlib Axes object. If None, a new figure is created.
        """
        import matplotlib.pyplot as plt
        
        elements = list(self.curves.keys())

        for e in elements:
            X = self.curves[e].peak_areas
            Y = self.curves[e].concentrations
            pred = self.curves[e].prediction
            
            plt.scatter(X, Y)
            plt.plot(X, pred, color='C1')
            plt.title(f'{e} MSE = {self.curves[e].mse:.2f}')
            plt.xlabel('Peak Area (counts)')
            plt.ylabel('Concentration')
            plt.show()

    def run_regression(self):
        """
        Performs linear regression on the calibration data for each element to determine
        the relationship between peak areas and concentrations. The slope (m), intercept (b), 
        mean squared error (MSE), and R-squared value are computed for each element.
        """
        elements = list(self.curves.keys())
        for e in elements:
            peak_areas = self.curves[e].peak_areas
            X = array(peak_areas).reshape(-1, 1)
            concs = self.curves[e].concentrations
            y = array(concs)

            regr = linear_model.LinearRegression(fit_intercept=True)
            regr.fit(X, y)

            y_pred = regr.predict(X)

            self.curves[e].lm = regr
            self.curves[e].b = regr.intercept_
            self.curves[e].m = regr.coef_[0]
            self.curves[e].mse = mean_squared_error(y, y_pred)
            self.curves[e].r2 = r2_score(y, y_pred)
            self.curves[e].prediction = y_pred


from lcicpms.integrate import Integrate
from numpy import array 

def get_peak_start_end_times(time_range : tuple = (-1, -1), all_times : array = None):

    if time_range[0] == -1:
        tmin = min(all_times)
    else:
        tmin = time_range[0]

    if time_range[1] == -1:
        tmax = max(all_times)
    else:
        tmax = time_range[1]

    return (tmin, tmax)


def get_peak_area(intensities : dict = None, times : dict = None, time_range : tuple = (-1,-1)):

    trange = get_peak_start_end_times(time_range, times)

    area = Integrate.integrate(intensities, times, trange)

    return area 


'''
@author: Christian Dewey
@date: Jul 27, 2024
'''
from numpy import where

class Integrate:
    """
    A class to handle integration of signal traces over a given time range.
    """

    @staticmethod
    def integrate(intensities, times, time_range: tuple = None):
        """
        Computes the area under the signal trace curve between a specified time range.

        This function performs trapezoidal integration on the provided intensities and times,
        returning the total area within the time range. If no time range is provided, it will
        integrate over the entire signal.

        Args:
            intensities (array-like): The intensities corresponding to each time point.
            times (array-like): The time points at which the intensities were measured.
            time_range (tuple, optional): The start and end times for the integration. 
                                          If None, the entire signal is integrated. Defaults to None.

        Returns:
            float: The computed area under the curve between the specified time range.
        """
        if time_range is None:
            i_tmin = 0
            i_tmax = len(times) - 2
        else:
            i_tmin = int(where(abs(times - time_range[0]) == min(abs(times - time_range[0])))[0][0])
            i_tmax = int(where(abs(times - time_range[1]) == min(abs(times - time_range[1])))[0][0])

        peak_area = 0
        dt = 0

        for i in range(i_tmin, i_tmax):
            dt1_intensities = intensities[i]
            dt2_intensities = intensities[i + 1]
            
            dt_I_min = min([dt1_intensities, dt2_intensities])
            dt_I_max = max([dt1_intensities, dt2_intensities])
            
            dt = (times[i + 1] - times[i])

            rect_area = dt * dt_I_min
            top_area = dt * (dt_I_max - dt_I_min) * 0.5
            d_peak_area = rect_area + top_area

            peak_area += d_peak_area

        return peak_area


'''
@author: Christian Dewey
@date: Jul 27, 2024
'''

from lcicpms.raw_icpms_data import RawICPMSData
from lcicpms.calibration import Calibration
from lcicpms.in_out import export_df
from lcicpms.integrate import Integrate
import lcicpms.utils as utils

class Quantitate:
    """
    A class to perform quantitation of elements using LC-ICP-MS data. It applies
    calibration curves to raw LC-ICP-MS data to compute concentrations over a given time range.
    """

    def run(raw_icpms_data: RawICPMSData = None, 
            calibration: Calibration = None,
            elements: list = [], 
            time_range: tuple = (-1, -1), 
            istd_baseline: float = -1.0, 
            istd_time_range: tuple = (-1, -1),
            baseline_subtraction: bool = False, 
            unit: str = 'ppb'):
        """
        Perform quantitation for a specified list of elements over a time range.

        Args:
            raw_icpms_data (RawICPMSData): An instance of RawICPMSData containing 
                                           raw ICP-MS intensities and times.
            calibration (Calibration): An instance of Calibration containing 
                                       calibration curves for elements.
            elements (list): A list of element names (strings) to be quantified.
            time_range (tuple): A tuple representing the start and end times 
                                (in seconds) for quantitation. Defaults to (-1, -1),
                                which indicates using the entire time range.
            istd_baseline (float): The baseline value for internal standard correction. 
                                   If > 0, internal standard correction is applied.
                                   Defaults to -1.0 (no correction).
            istd_time_range (tuple): A tuple representing the time range for the 
                                     internal standard. Defaults to (-1, -1).
            baseline_subtraction (bool): If True, baseline subtraction is applied. 
                                         Defaults to False.
            unit (str): The unit of the concentration (e.g., 'ppb'). Defaults to 'ppb'.

        Returns:
            dict: A dictionary containing:
                - 'areas': DataFrame of integrated peak areas for each element.
                - 'concentrations': DataFrame of concentrations calculated for each element.
                - 'internal_std_correction': The internal standard correction factor applied.
        
        """
        
        intensities = raw_icpms_data.intensities
        times = raw_icpms_data.times

        istd_correction = 1 

        if istd_baseline > 0:
            istd_trange = utils.get_peak_start_end_times(istd_time_range, times['115In'])
            istd_area = Integrate.integrate(intensities['115In'], times['115In'], istd_trange)
            istd_correction = istd_baseline / istd_area
            

        peak_areas = {}
        concentrations = {}
        trange = [-1,-1]

        for element in elements:

            trange = utils.get_peak_start_end_times(time_range, times[element])
            area = Integrate.integrate(intensities[element], times[element], trange)

            peak_areas[element] =  float(area) * istd_correction

            calibration_curve = calibration.curves[element]
            m =  calibration_curve.m
            b = calibration_curve.b
            concentrations[element] = float(m * area * istd_correction + b )
        
        return {'areas':export_df(peak_areas, trange), 'concentrations':export_df(concentrations, trange), 'internal_std_correction':istd_correction}


'''
@author: Christian Dewey
@date: Jul 27, 2024
'''

from pandas import read_csv, DataFrame

def get_data_from_csv(path_to_rawfile=None):
    """
    Reads a raw ICP-MS data file from a CSV format and returns a pandas DataFrame.
    
    The function automatically detects the header row by searching for the 'Time' keyword.
    It uses this information to appropriately parse the file starting from the correct row.

    Args:
        path_to_rawfile (str, optional): The file path to the raw data CSV. Defaults to None.

    Returns:
        DataFrame: A pandas DataFrame containing the loaded raw data.
    """
    # Start by finding the header line
    df = read_csv(path_to_rawfile, sep=None, nrows=5, engine="python")
    if any("Time" in c for c in list(df.columns)):
        istart = 0
    else:
        istart = 1
        for row in df.itertuples(): 
            if row[1].find('Time') != -1:
                break
            istart += 1 
    df = read_csv(path_to_rawfile, sep=None, header=istart, engine="python")
    return df


def export_df(data: dict, time_range: tuple):
    """
    Exports data to a pandas DataFrame, including the start and end of a time range.
    
    The function takes in a dictionary of data and a time range, adding the start 
    and end times as columns. It ensures that the final DataFrame has 'start' and 
    'end' columns followed by the other data columns.

    Args:
        data (dict): A dictionary where keys are element names and values are the corresponding data.
        time_range (tuple): A tuple representing the start and end times of the time range.

    Returns:
        DataFrame: A pandas DataFrame with 'start', 'end', and element data columns.
    """
    columns = ['start', 'end'] + [e for e in data.keys()]
    data['start'] = float(time_range[0])
    data['end'] = float(time_range[1])
    for k in columns:
        data[k] = [data[k]]
    df = DataFrame(data, columns=data.keys())
    return df[columns]